package industries.sunshine.planningpoker

import industries.sunshine.planningpoker.common.Models.*
import cats.effect.{Ref, Sync}
import cats.syntax.all._

enum RoomError {
  case RoomAlreadyExists(name: String)
}

trait RoomService[F[_]] {
  def createRoom(newRoom: Room): F[Either[RoomError, Room]]
  def updateRoom(room: Room): F[Unit]
  def deleteRoom(roomID: RoomID): F[Unit]
  def getRoom(roomID: RoomID): F[Option[Room]]
}

class InMemoryRoomService[F[_]: Sync](stateRef: Ref[F, Map[RoomID, Room]])
    extends RoomService[F] {
  override def createRoom(newRoom: Room): F[Either[RoomError, Room]] = {
    stateRef.modify { rooms =>
      rooms.get(newRoom.id) match {
        case Some(_) =>
          rooms -> RoomError.RoomAlreadyExists(newRoom.id.name).asLeft[Room]
        case None =>
          (rooms.updated(newRoom.id, newRoom)) -> newRoom.asRight[RoomError]
      }
    }
  }
  def updateRoom(room: Room): F[Unit] = stateRef.update { state =>
    state.get(room.id).fold(state)(oldRoom => state.updated(room.id, room))
  }

  def deleteRoom(roomID: RoomID): F[Unit] = stateRef.update(_.removed(roomID))

  def getRoom(roomID: RoomID): F[Option[Room]] = stateRef.get.map(_.get(roomID))

}
object RoomService {
  def make[F[_]: Sync]: F[RoomService[F]] = {
    Ref.of[F, Map[RoomID, Room]](Map.empty).map(new InMemoryRoomService[F](_))
  }
}
